/*

    Program: reflector2midi - tool to be used with asciireflector to
    convert text to MIDI control events.

    Copyright (C) 1999/2000  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try:
    |  Andreas Schiffler                    aschiffler@home.com  |
    |  4707 Eastwood Cres., Niagara Falls, Ont  L2E 1B4, Canada  |
    |  +1-905-371-3652 (private)  -  +1-905-371-8834 (work/fax)  |

*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <sys/time.h>

#include <netinet/in.h>

#include <limits.h>
#include <netdb.h>
#include <arpa/inet.h>

#include <sys/socket.h>
#include <errno.h>

#include <pthread.h> 

/* Text buffer sizes */

#define LINE_LEN	256
#define BUFFER_LEN	128

/* ----- Configurable items ----- */

unsigned short PORT;
char HOSTNAME[LINE_LEN];
int RATE;
char MIDIDEV[LINE_LEN];
char CONFFILE[LINE_LEN];
char LOGFILE[LINE_LEN];
int CHANNEL;

/* ------- Event definitions -------- */

/* Event count */

int event_count=0;

/* ASCII character for event types */

#define EVENT_TYPE_TRIGGER	'T'
#define EVENT_TYPE_COUNTER	'C'
#define EVENT_TYPE_VALUE	'V'

/* Event structure for event list */

struct tevent {
		char type;
		int control;
		int value;
		char name[LINE_LEN];
		int namelen;
		char pattern[LINE_LEN];
		int patternlen;
		int updated;
		void *next;
	      };

struct tevent *rootevent=NULL,*curevent=NULL;

/* =========== Subroutines ============= */

/* ------- Configuration routines  ------- */

void set_defaults()
{
 strcpy (HOSTNAME,"localhost");
 PORT=1522;
 strcpy (MIDIDEV,"/dev/midi00");
 RATE=10;
 CHANNEL=3;
 strcpy(CONFFILE,"events.cfg");
 strcpy(LOGFILE,"events.log");
}

/* -------- Event routines ---------- */

/* Add an event to the list */

void addEvent(char *description, FILE *logfile)
{
 struct tevent *thisevent;
  
 /* Create entry */
 thisevent=(struct tevent *)malloc(sizeof(struct tevent));

 /* Parse description */
 sscanf(description,"%c %i %s %s",&thisevent->type,&thisevent->control,thisevent->name,thisevent->pattern);
 thisevent->namelen=strlen(thisevent->name);
 thisevent->patternlen=strlen(thisevent->pattern);
 
 /* Message for debugging purposes*/
 fprintf (logfile,"Event: TYPE=%c CONTROL=%i SOURCE=%s MATCH=%s added.\n",thisevent->type,thisevent->control,thisevent->name,thisevent->pattern);
 
 /* Link entry into list */
 if (rootevent==NULL) {
  rootevent=thisevent;
  curevent=thisevent;
 } else {
  curevent->next=thisevent;
  curevent=thisevent;
 }    
 
 /* Count */
 event_count++;
}

/* Load event configuration data */

int loadEventConfiguration()
{
 FILE *infile=NULL;
 FILE *logfile=NULL;
 char aline[LINE_LEN];

 /* Reset count */
 event_count=0;
  
 /* Open files */
 if ((infile=fopen(CONFFILE,"r"))==NULL) {
  perror ("fopen()");
  return(-1);
 } else {
  if ((logfile=fopen(LOGFILE,"w"))==NULL) {
   perror ("fopen()");
   fclose(infile);
   return(-1);
  } else {
   /* Read all lines in the configuration file */
   fgets(aline,LINE_LEN,infile);
   while (!feof(infile)) {
    /* Skip comment lines */
    if ((aline[0] != '#') && (aline[0] != '\n')) {
     addEvent(aline,logfile);
    } 
    fgets(aline,LINE_LEN,infile);
   }
  } 
 }

 /* Close files */  
 fclose(infile);
 fclose(logfile);

 return(0);
}

/* ---- Network routines ------- */

/* Converts ASCII text to in_addr struct */

struct in_addr *atoaddr(address)
char *address;
{
  struct hostent *host;
  static struct in_addr saddr;

  /* First try it as aaa.bbb.ccc.ddd. */
  saddr.s_addr = inet_addr(address);
  if (saddr.s_addr != -1) {
    return &saddr;
  }
    
  host = gethostbyname(address);
  if (host != NULL) {
    return (struct in_addr *) *host->h_addr_list;
  }
  return NULL;
}

/* Write buffer to socket connection */
   
int sock_write (int sockfd, char *buf, size_t count)
{
 size_t bytes_sent = 0;
 int this_write;
       
 while (bytes_sent < count) {
  do {
   this_write = write(sockfd, buf, count - bytes_sent);
  } while ( (this_write < 0) && (errno == EINTR) );
  if (this_write <= 0) {
   return(this_write);
  } 
  bytes_sent += this_write;
  buf += this_write;
 }
 return(count);
}

/* Read from socket connection into buffer */

int sock_read (int sockfd, char *buf, size_t count)
{
 size_t bytes_read = 0;
 int this_read;
 fd_set readers;
 struct timeval timeout;
 int pending;

 while (bytes_read < count) {
  do {
   /* Check status of connection */
   pending=0;
   while (pending==0) {
    FD_ZERO(&readers);
    FD_SET(sockfd,&readers);
    timeout.tv_sec=5;
    timeout.tv_usec=0;
    pending=select(FD_SETSIZE,&readers,0,0,&timeout);
    if (pending<0) {
     fprintf (stderr,"Connection closed ...\n");
     return(-1);
    }
   } 
   this_read = read(sockfd, buf, count - bytes_read);
  } while ( (this_read < 0) && (errno == EINTR) );
  if ((this_read==0) && (pending==1)) {
   fprintf (stderr,"Connection closed ...\n");
   return(-1);
  }    
  if (this_read <= 0) {
   return(this_read);
  } 
  bytes_read += this_read;
  buf += this_read;
 }
 return(count);
}

/* Read a string delimited by a newline character from a socket connection */
      
int sock_gets (int sockfd, char *str, size_t count)
{
 int bytes_read;
 int total_count = 0;
 char *current_position;
 char last_read = 0;
 fd_set readers;
 struct timeval timeout;
 int pending;
          
 current_position = str;
 while ((last_read != 10) && (last_read !=13)) {
   /* Check status of connection */
   pending=0;
   while (pending==0) {
    FD_ZERO(&readers);
    FD_SET(sockfd,&readers);
    timeout.tv_sec=5;
    timeout.tv_usec=0;
    pending=select(FD_SETSIZE,&readers,0,0,&timeout);
    if (pending<0) {
     fprintf (stderr,"Connection closed ...\n");
     return(-1);
    }
   } 
   bytes_read = read(sockfd, &last_read, 1);
   if ((bytes_read==0) && (pending==1)) {
    fprintf (stderr,"Connection closed ...\n");
    return(-1);
   }    
   if (bytes_read <= 0) {
    return bytes_read; 
   }
   if ( (total_count < count) && (last_read != 10) && (last_read !=13) ) {
    current_position[0] = last_read;
    current_position++;
    total_count++;
   }
 }
 if (count > 0) current_position[0] = 0;
 return total_count;
}

/* ---------- MIDI routines --------- */

/* Send a MIDI control byte */

int midiSendControl(int midi_fd,unsigned char channel,unsigned char control,unsigned char value)
{
 unsigned char message[3];

#ifdef DEBUG
 fprintf (stderr,"Channel=%i Control=%i Value=%i\n",channel,control,value); 
#endif
 
 message[0] = 0xb0 |(channel & 0x0f);
 message[1] = control & 0x7f;
 message[2] = value   & 0x7f;

 if (write (midi_fd,message,3)!=3) {
  perror("write()");
  return(-1); 
 } else {
  return(0);
 }   
}

/* ======== Main processing loop ============= */

void NetworkLoop(int connected_socket, int midi_fd)
{
 int counter=0;
 int connected;
 int readloop;
 int numread;
 char outbuffer[BUFFER_LEN];
 char inbuffer[BUFFER_LEN];
 char *found=NULL,*haystack=NULL;
 struct tevent *thisevent;
 float fvalue;
 int USLEEP;
 
 /* Calculate rate-throttle pause */
 USLEEP=1000000/RATE;  
 /* Processing loop */
 connected=1;
 while (connected) {
  /* Count iterations */
  counter++;
  /* Reset event values and update flag */
  thisevent=rootevent;
  while (thisevent!=NULL) {
   thisevent->value=0;
   thisevent->updated=0;
   thisevent=(struct tevent *)thisevent->next;
  } 
  /* Write POLL to reflector */
  sprintf(outbuffer,"POLL\n");
  connected=(sock_write(connected_socket,outbuffer,strlen(outbuffer))>0);
  /* Read text until OVER has been reached */
  readloop=1;
  while (readloop) {
   if ( (numread=sock_gets(connected_socket,inbuffer,BUFFER_LEN))>0) {
    /* Parse textlines */
    if (strncasecmp(inbuffer,"OVER",4)==0) {
     readloop=0;
    } else {
     /* Process inbuffer */
     thisevent=rootevent;
     while (thisevent!= NULL) {
      /* First compare name */
      if (strncasecmp(inbuffer,thisevent->name,thisevent->namelen)==0) {
       switch (thisevent->type) {
        case EVENT_TYPE_TRIGGER:
         /* Match ... then look for pattern */
         haystack=inbuffer;
         found=strstr(haystack,thisevent->pattern);
         if (found != NULL) {
          thisevent->value |= 1;
          thisevent->updated = 1;
         }
         break; 
        case EVENT_TYPE_COUNTER:
         /* Match ... then count for patterns */
         haystack=inbuffer;
         found=strstr(haystack,thisevent->pattern);
         while (found != NULL) {
          if (found != NULL) {
           thisevent->value++;
           thisevent->updated = 1;
          } 
          found += thisevent->patternlen;
          found=strstr(found,thisevent->pattern);
         } 
         break; 
        case EVENT_TYPE_VALUE:
         /* Match ... then look for pattern and convert to value */
         haystack=inbuffer;
         found=strstr(haystack,thisevent->pattern); 
         if (found!=NULL) {
          /* Adjust for patternlength */
          found += thisevent->patternlen;
          /* Get value */
          fvalue=atof(found);
          /* Adjust range */
          if ((fvalue>=0.0) && (fvalue<=1.0)) {
           thisevent->value=(unsigned char)(fvalue*127.0);
          } else if ((fvalue>=0.0) && (fvalue<=127.0)) {
           thisevent->value=(unsigned char)(fvalue);
          } else if (fvalue<0.0) {
           thisevent->value=0;
          } else {
           thisevent->value=127;
          }
          thisevent->updated = 1;
         }
         break; 
       } 
      }
      thisevent=(struct tevent *)thisevent->next;
     }
    }
   } else {
    readloop=0;
   } 
  }
  /* Send MIDI data based on events */
  thisevent=rootevent;
  while (thisevent!=NULL) {
   if (thisevent->updated==1) {
    midiSendControl(midi_fd,CHANNEL,thisevent->control,thisevent->value);
   }
   thisevent=(struct tevent *)thisevent->next;
  } 
  /* Pause */
  usleep(USLEEP);
 }
}

/* Main routine */

int main (int argc, char *argv[])
{
 int i;
 struct sockaddr_in address;
 static struct in_addr *addr;
 int connected;
 int connected_socket = -1;
 int midi_fd;

 /* Print message */  
 printf ("reflector2midi - A. Schiffler, 08/1999-04/2000\n\n");

 /* Initialize variables */
 set_defaults();
 
 /* Parse command line */
 for (i=1; i<argc; i++) {
  char	*arg;
  if (argv[i][0]!='-') break;
  arg = argv[i];
  arg++;
  if (*arg == '-') arg++;
  if (strncmp(arg,"port",4)==0) {
   i++; 
   if (i<=argc) {
    PORT = atoi(argv[i]);
   }
  } else if (strncmp(arg,"rate",4)==0) {
   i++; 
   if (i<=argc) {
    RATE = atoi(argv[i]);
   }
  } else if (strncmp(arg,"channel",7)==0) {
   i++; 
   if (i<=argc) {
    CHANNEL = atoi(argv[i]);
   }
  } else if (strncmp(arg,"mididev",7)==0) {
   i++; 
   if (i<=argc) {
    strcpy (MIDIDEV,argv[i]);
   }
  } else if (strncmp(arg,"host",4)==0) {
   i++; 
   if (i<=argc) {
    strcpy (HOSTNAME,argv[i]);
   }
  } else if (strncmp(arg,"config",6)==0) {
   i++; 
   if (i<=argc) {
    strcpy (CONFFILE,argv[i]);
   }
  } else if (strncmp(arg,"log",3)==0) {
   i++; 
   if (i<=argc) {
    strcpy (LOGFILE,argv[i]);
   }
  } else {
   /* Command line help */
   printf ("Command line parameters:\n");
   printf ("\nNetwork settings:\n");
   printf (" -host name      host to connect to (default: %s)\n",HOSTNAME);
   printf (" -port #         port to connect to (default: %i)\n",PORT);
   printf ("\nConfiguration settings:\n");
   printf (" -config name    event configuration filename (default: %s)\n",CONFFILE);
   printf (" -log name       configuration log filename (default: %s)\n",LOGFILE);
   printf ("\nOutput settings:\n");
   printf (" -mididev dev    midi device for output (default: %s)\n",MIDIDEV);
   printf (" -rate #         update rate for scanning reflector in Hz (default: %i)\n",RATE);
   printf (" -channel #      midi channel to output on (default: %i)\n",CHANNEL);
   exit(0);
  }
 }
 
 /* Load event config */
 printf ("Loading event configuration ... \n");
 if (loadEventConfiguration()<0) {
  printf ("\n");
  fprintf(stderr,"Error processing %s.\n",CONFFILE);
  exit(-1);
 }
 printf ("%i Events loaded.\n",event_count);
 
 /* Setup internet address information */
 addr = (struct in_addr *)atoaddr((char *)HOSTNAME);
 if (addr == NULL) {
  fprintf(stderr,"Invalid network address: %s",HOSTNAME);
  exit(-1);
 }
 memset((char *) &address, 0, sizeof(address));
 address.sin_family = AF_INET;
 address.sin_addr.s_addr = addr->s_addr;
 address.sin_port = htons(PORT);

 /* Connect to reflector */ 
 connected_socket = socket(AF_INET, SOCK_STREAM, 0);
 printf("Connecting to %s on port %d ...",inet_ntoa(*addr),PORT);
 connected = connect(connected_socket, (struct sockaddr *) &address, sizeof(address));
 if (connected < 0) {
  printf("\n");
  perror("connect()");
  exit(-1);
 }
 printf(" connected.\n");

 /* Open MIDI port */
 printf("Opening MIDI device %s on channel %i.\n",MIDIDEV,CHANNEL);
 if ((midi_fd=open(MIDIDEV,O_WRONLY))<0) {
  perror("open()");
  close(connected_socket);
  exit(-1);
 }
  
 /* Start processing the reflector text and generating MIDI controls */
 printf ("Processing ...\n");
 NetworkLoop(connected_socket,midi_fd);
 
 /* Close file/socket */
 close(connected_socket);
 close(midi_fd);
              
 exit(0);
}

